#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "ExifHelper.h"

#define EXIF_SUPPORT_GPS        (1)

namespace android_camera_hal3 {

// Exif helper functions

// alloc format ref: https://github.com/LuaDist/im/blob/master/src/im_format_jpeg.cpp
// ref http://www.opanda.com/en/pe/help/gps.html
// ref https://github.com/LuaDist/im/blob/master/src/im_format_jpeg.cpp
// ref https://github.com/rnorris/viking/blob/master/src/geotag_exif.c
// ref http://fossies.org/linux/viking/src/geotag_exif.c
ExifEntry* my_exif_create_value (ExifData *ed, ExifTag tag, ExifIfd ifd)
{
  ExifEntry *e = exif_content_get_entry (ed->ifd[ifd], tag);
  if ( !e ) {
      e = exif_entry_new ();
      exif_content_add_entry (ed->ifd[ifd], e);

      exif_entry_initialize (e, tag);

      // exif_entry_initialize doesn't seem to do much, especially for the GPS tags
      // so have to setup fields ourselves:
      e->tag = tag;

      if ( tag == EXIF_TAG_GPS_VERSION_ID ) {
          e->format = EXIF_FORMAT_BYTE;
          e->components = 4;
          e->size = sizeof (ExifByte) * e->components;
          if ( e->data )
              free (e->data);
          e->data = (unsigned char *)malloc (e->size);
      }
      if ( tag == EXIF_TAG_GPS_MAP_DATUM ||
           tag == EXIF_TAG_GPS_LATITUDE_REF ||
           tag == EXIF_TAG_GPS_LONGITUDE_REF ||
           tag == EXIF_TAG_GPS_PROCESSING_METHOD) {
          e->format = EXIF_FORMAT_ASCII;
          // NB Allocation is handled later on when the actual string used is known
      }

      if ( tag == EXIF_TAG_GPS_LATITUDE || tag == EXIF_TAG_GPS_LONGITUDE ) {
          e->format = EXIF_FORMAT_RATIONAL;
          e->components = 3;
          e->size = sizeof (ExifRational) * e->components;
          if ( e->data )
              free (e->data);
          e->data = (unsigned char *)malloc (e->size);
      }
      if ( tag == EXIF_TAG_GPS_ALTITUDE ) {
          e->format = EXIF_FORMAT_RATIONAL;
          e->components = 1;
          e->size = sizeof (ExifRational) * e->components;
          if ( e->data )
              free (e->data);
          e->data = (unsigned char *)malloc (e->size);
      }
      if ( tag == EXIF_TAG_GPS_ALTITUDE_REF ) {
          // assign a type
          e->format = EXIF_FORMAT_BYTE;
          e->components = 1;
          e->size = sizeof (ExifByte) * e->components;
          if ( e->data )
              free (e->data);
          e->data = (unsigned char *)malloc (e->size);
      }
      if ( tag == EXIF_TAG_GPS_DATE_STAMP ) {
            e->format = EXIF_FORMAT_ASCII;
            e->components = 11;
            e->size = sizeof (ExifAscii) * e->components;
            if ( e->data )
                free (e->data);
            e->data = (unsigned char *)malloc (e->size);
        }
      if ( tag == EXIF_TAG_GPS_TIME_STAMP ) {
          e->format = EXIF_FORMAT_RATIONAL;
          e->components = 3;
          e->size = sizeof (ExifRational) * e->components;
          if ( e->data )
              free (e->data);
          e->data = (unsigned char *)malloc (e->size);
      }
      /* The entry has been added to the IFD, so we can unref it */
      //exif_entry_unref(e);
      // Crashes later on, when saving to jpeg if the above unref is enabled!!
      // ?Some other malloc problem somewhere?
  }
  return e;
}

void convert_to_entry (const char *set_value, double gdvalue, ExifEntry *e, ExifByteOrder o)
{
    unsigned int i, numcomponents;
    char *value_p = NULL;
    char *buf = NULL;
    /*
    * ASCII strings are handled separately,
    * since they don't require any conversion.
    */
    if (e->format == EXIF_FORMAT_ASCII || e->tag == EXIF_TAG_USER_COMMENT) {

      if (e->data) free (e->data);

      e->components = strlen (set_value) + 1;
      // Important! cts testing GpsProcessingMethod setting
      if (e->tag == EXIF_TAG_USER_COMMENT || e->tag==EXIF_TAG_GPS_PROCESSING_METHOD)
          e->components += 8 - 1;
      e->size = sizeof (char) * e->components;
      e->data = (unsigned char *)malloc (e->size);
      if (!e->data) {
          //g_warning (_("Not enough memory."));
          return;
      }
      // Important! cts testing GpsProcessingMethod setting
      if (e->tag == EXIF_TAG_USER_COMMENT || e->tag==EXIF_TAG_GPS_PROCESSING_METHOD) {
          /* assume ASCII charset */
          /* TODO: get this from the current locale */
          memcpy ((char *) e->data, "ASCII\0\0\0", 8);
          memcpy ((char *) e->data + 8, set_value,
          strlen (set_value));
      } else
          strncpy ((char *) e->data, set_value, e->size);
          return;
    }

    /*
    * Make sure we can handle this entry
    */
    if ((e->components == 0) && *set_value) {
      //g_warning (_("Setting a value for this tag is unsupported!"));
      return;
    }

    bool use_string = (set_value != NULL);
    if ( use_string ) {
      /* Copy the string so we can modify it */
      buf = strdup (set_value);
      if (!buf)
          return;
      value_p = strtok (buf, " ");
    }

    numcomponents = e->components;
    for (i = 0; i < numcomponents; ++i) {
          unsigned char s;

        if ( use_string ) {
          if (!value_p) {
              //g_warning (_("Too few components specified (need %d, found %d)\n"), numcomponents, i);
              return;
          }
          if (!isdigit(*value_p) && (*value_p != '+') && (*value_p != '-')) {
              //g_warning (_("Numeric value expected\n"));
              return;
          }
        }

        s = exif_format_get_size (e->format);
        switch (e->format) {
            case EXIF_FORMAT_ASCII:
               //g_warning (_("This shouldn't happen!"));
               return;
               break;
            case EXIF_FORMAT_SHORT:
               exif_set_short (e->data + (s * i), o, atoi (value_p));
               break;
            case EXIF_FORMAT_SSHORT:
               exif_set_sshort (e->data + (s * i), o, atoi (value_p));
               break;
            case EXIF_FORMAT_RATIONAL: {
                  ExifRational er;

                  double val = 0.0 ;
                  if ( use_string && value_p )
                       val = fabs (atof (value_p));
                  else val = fabs (gdvalue);

                  if ( i == 0 ) {
                  // One (or first) part rational

                      // Sneak peek into tag as location tags need rounding down to give just the degrees part
                      if ( e->tag == EXIF_TAG_GPS_LATITUDE || e->tag == EXIF_TAG_GPS_LONGITUDE ) {
                          er.numerator = (ExifLong) floor ( val );
                          er.denominator = 1.0;
                      }
                      else {
                          // I don't see any point in doing anything too complicated here,
                          // such as trying to work out the 'best' denominator
                          // For the moment use KISS principle.
                          // Fix a precision of 1/100 metre as that's more than enough for GPS accuracy especially altitudes!
                          er.denominator = 100.0;
                          er.numerator = (ExifLong) (val * er.denominator);
                      }
                  }

                  // Now for Location 3 part rationals do Mins and Seconds format

                  // Rounded down minutes
                  if ( i == 1 ) {
                      er.denominator = 1.0;
                      er.numerator = (ExifLong) ( (int) floor ( ( val - floor (val) ) * 60.0 ) );
                  }

                  // Finally seconds
                  if ( i == 2 ) {
                      er.denominator = 100.0;
                      // Fractional minute.
                      double FracPart = ((val - floor(val)) * 60) - (double)(int) floor ( ( val - floor (val) ) * 60.0 );
                      er.numerator = (ExifLong) ( (int)floor(FracPart * 6000) ); // Convert to seconds.
                  }
                  exif_set_rational (e->data + (s * i), o, er );
                  break;
                  }
            case EXIF_FORMAT_LONG:
            exif_set_long (e->data + (s * i), o, atol (value_p));
            break;
            case EXIF_FORMAT_SLONG:
            exif_set_slong (e->data + (s * i), o, atol (value_p));
            break;
            case EXIF_FORMAT_BYTE:
            case EXIF_FORMAT_SBYTE:
            case EXIF_FORMAT_UNDEFINED: /* treat as byte array */
            e->data[s * i] = atoi (value_p);
            break;
            case EXIF_FORMAT_FLOAT:
            case EXIF_FORMAT_DOUBLE:
            case EXIF_FORMAT_SRATIONAL:
            default:
            //g_warning (_("Not yet implemented!"));
            return;
        }

        if ( use_string )
            value_p = strtok (NULL, " ");
    }

  free (buf);
}

bool add_exif(JPEG_RST_INFO *pInfo,
        uint8_t *metaBuf,
        /* , uint32_t bufSize */
        MISC_FIELD_INFO *pMiscInfo)
{
    memcpy(metaBuf,pInfo->vaddr,pInfo->jpegSize);
    JPEGData *pJpgData = jpeg_data_new_from_data(
                (const unsigned char*)metaBuf,
                pInfo->jpegSize);

    if(pJpgData == NULL){
        ALOGE("%s %d return error",__FUNCTION__,__LINE__);
        return false;
    }

    // Get current values
    ExifData *ed = exif_data_new_from_data(
                (const unsigned char*)metaBuf,
                pInfo->jpegSize);

    if ( !ed )
        ed = exif_data_new ();
    ExifEntry *ee;

    // EXIF_IFD_0 section
    ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_MAKE , EXIF_IFD_0);
    convert_to_entry ( "Realtek", 0.0, ee, exif_data_get_byte_order(ed) );
    ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_MODEL , EXIF_IFD_0);
    convert_to_entry ( "Realtek", 0.0, ee, exif_data_get_byte_order(ed) );
    ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_FOCAL_LENGTH, EXIF_IFD_EXIF);
    // some value is fixed , which is not good however...
    convert_to_entry ( "5.0", 0.0, ee, exif_data_get_byte_order(ed) );

    // EXIF spec is "YYYY:MM:DD HH:MM:SS".
    time_t _time;
    time(&_time);
    struct tm* tstruct;
    tstruct = localtime(&_time);
    char szCurrentDateTime[100];
    sprintf(szCurrentDateTime,
        "%04d:%02d:%02d %02d:%02d:%02d",
        tstruct->tm_year + 1900,
        tstruct->tm_mon + 1,
        tstruct->tm_mday,
        tstruct->tm_hour,
        tstruct->tm_min,
        tstruct->tm_sec);

    ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_DATE_TIME, EXIF_IFD_0);
    convert_to_entry ( szCurrentDateTime, 0.0, ee, exif_data_get_byte_order(ed) );

#if (EXIF_SUPPORT_GPS)
    double mExifGpsLatitude = pMiscInfo->gpsCoordinates[0];
    if(mExifGpsLatitude != 0.0){
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_LATITUDE, EXIF_IFD_GPS);
        convert_to_entry ( NULL, mExifGpsLatitude, ee, exif_data_get_byte_order(ed) );
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_LATITUDE_REF, EXIF_IFD_GPS);
        convert_to_entry ( mExifGpsLatitude < 0.0 ? "S" : "N", 0.0, ee, exif_data_get_byte_order(ed) );
    }
    double mExifGpsLongitude = pMiscInfo->gpsCoordinates[1];
    if(mExifGpsLongitude != 0.0){
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_LONGITUDE, EXIF_IFD_GPS);
        convert_to_entry ( NULL, mExifGpsLongitude, ee, exif_data_get_byte_order(ed) );
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_LONGITUDE_REF, EXIF_IFD_GPS);
        convert_to_entry ( mExifGpsLongitude < 0.0 ? "W" : "E", 0.0, ee, exif_data_get_byte_order(ed) );
    }
    double mExifGpsAltitude = pMiscInfo->gpsCoordinates[2];
    if(mExifGpsAltitude != 0.0){
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_ALTITUDE, EXIF_IFD_GPS);
        convert_to_entry ( NULL, mExifGpsAltitude, ee, exif_data_get_byte_order(ed) );
        // byte 0 meaning "sea level" or 1 if the value is negative.
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_ALTITUDE_REF, EXIF_IFD_GPS);
        convert_to_entry ( mExifGpsAltitude < 0.0 ? "1" : "0", 0.0, ee, exif_data_get_byte_order(ed) );
    }

    int mExifGpsTimeStampYear = -1;
    int mExifGpsTimeStampMonth = -1;
    int mExifGpsTimeStampDay = -1;
    int mExifGpsTimeStampHour = -1;
    int mExifGpsTimeStampMin = -1;
    int mExifGpsTimeStampSec = -1;

    if(pMiscInfo->gpsTimestamp > 0) {
        struct tm *timeinfo = gmtime( ( time_t * ) & (pMiscInfo->gpsTimestamp) );
        mExifGpsTimeStampYear = timeinfo->tm_year;
        mExifGpsTimeStampMonth = timeinfo->tm_mon;
        mExifGpsTimeStampDay = timeinfo->tm_mday;
        mExifGpsTimeStampHour = timeinfo->tm_hour;
        mExifGpsTimeStampMin = timeinfo->tm_min;
        mExifGpsTimeStampSec = timeinfo->tm_sec;
    }

    if((mExifGpsTimeStampYear>-1) && (mExifGpsTimeStampMonth>-1) && (mExifGpsTimeStampDay>-1)){
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_DATE_STAMP, EXIF_IFD_GPS);
        sprintf(szCurrentDateTime,
            "%04d:%02d:%02d",
            mExifGpsTimeStampYear+1900,
            mExifGpsTimeStampMonth+1,
            mExifGpsTimeStampDay);
        convert_to_entry (szCurrentDateTime, 0.0, ee, exif_data_get_byte_order(ed) );
    }
    if((mExifGpsTimeStampHour>-1) && (mExifGpsTimeStampMin>-1) && (mExifGpsTimeStampSec>-1)){
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_TIME_STAMP, EXIF_IFD_GPS);
        sprintf(szCurrentDateTime,
            "%d %lf %lf",
            mExifGpsTimeStampHour,
            (double)mExifGpsTimeStampMin/60.0,
            (double)mExifGpsTimeStampSec/3600.0);
        convert_to_entry (szCurrentDateTime, 0.0, ee, exif_data_get_byte_order(ed) );

    }

    const char *mExifGpsProcessingMethod = (const char *)(pMiscInfo->gpsProcessingMethod);

    if(strlen(mExifGpsProcessingMethod)>0)
    {
        ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_PROCESSING_METHOD, EXIF_IFD_GPS);
        convert_to_entry (mExifGpsProcessingMethod,
            0.0,
            ee,
            exif_data_get_byte_order(ed) );
    }

    ee = my_exif_create_value (ed, (ExifTag)EXIF_TAG_GPS_VERSION_ID, EXIF_IFD_GPS);
    convert_to_entry ( "2 2 0 0", 0.0, ee, exif_data_get_byte_order(ed) );
#endif

    // add exif-thumbnail
    int iThumbnailWidth  = pMiscInfo->thumbWidth;
    int iThumbnailHeight = pMiscInfo->thumbHeight;
    if((iThumbnailWidth!=0) && (iThumbnailHeight!=0) && pInfo->pThumbBuf && pInfo->thumbSize)
    {
        ALOGV("[%s] add thumbnail in exif",__FUNCTION__);
        ed->data = (unsigned char*)pInfo->pThumbBuf;
        ed->size = pInfo->thumbSize;
    }

    jpeg_data_set_exif_data (pJpgData, ed);
    if(pJpgData) {
        unsigned char *d = NULL;
        unsigned int size = 0;
        jpeg_data_save_data (pJpgData, &d, &size);

        memset(pInfo->vaddr,0x0,pInfo->size);
        memcpy(pInfo->vaddr,d,size);

        free(d);
        jpeg_data_unref (pJpgData);

        ALOGV("%s complete",__FUNCTION__);
        return true;
    }

    ALOGE("%s %d return error",__FUNCTION__,__LINE__);
    return false;
}

}; // namespace android_camera_hal3
